from copy import deepcopy

from acelerai_pipelines.util import toUpperFirst

class EnvironmentItem:
    Name: str
    Value: str

    def __init__(self, **kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        return data


import asyncio
from datetime import datetime
import gzip
import hashlib
import json
import os
import pickle
import struct
import traceback
from acelerai_inputstream.models.asset import Asset
from acelerai_inputstream.models.pagination_data import PaginationData
import httpx
import math
import base64
from httpx import Response
import msgpack
import requests
from acelerai_inputstream import DATA_URL, APIGW, QUERY_MANAGER, VERIFY_HTTPS
from acelerai_inputstream.models.inputstream import INSERTION_MODE, ExternalDataConnectionDTO, Inputstream
from acelerai_inputstream.utils import CustomJSONEncoder, CustomJsonDecoder, custom_encoder, decode_datetime
import datetime

# Type conversion
from decimal import Decimal
from uuid import UUID

import logging
logger = logging.getLogger("HttpAcelerAI")

SEM = asyncio.Semaphore(20)  # Limitar concurrencia a 20 conexiones
MAX_RETRIES = 5  # Número máximo de reintentos
INITIAL_BACKOFF = 1  # Tiempo inicial de espera en segundos
BACKOFF_MULTIPLIER = 2  # Factor multiplicador para el tiempo de espera

packer = msgpack.Packer(default=custom_encoder) 

class AcelerAIHttpClient():

    def __init__(self, token:str):
        self.token = token
        self.lock = asyncio.Lock()
        
    async def execute_external_command(self, conn_key:str, query:str):
        """
        Execute an external command in the external data connection
        params:
            conn_key: str -> connection key
            query: str -> query/command to execute
        """    
        try:
            headers = {
                "Authorization": f"A2G {self.token}",
                "conn-key": conn_key,
                'Content-Type': 'application/json'
            }
            
            async with httpx.AsyncClient(verify=VERIFY_HTTPS) as client:
                res = await client.post(f"{QUERY_MANAGER}/QueryData/ExecuteExternalCommand", 
                    data=json.dumps(query, cls=CustomJSONEncoder),
                    headers=headers
                )
            
            if res.status_code != 200:
                if res.status_code == 404: raise Exception("ExternalDataConnection not found, please check your connkey")
                if res.status_code == 401: raise Exception("Unauthorized")
                if res.status_code == 403: raise Exception("Forbidden: please check your token or access permissions in the resource")
                raise Exception(f"Error executing external command, {res.status_code} {res.text}")
            
            content = res.json(cls=CustomJsonDecoder)
            
            if not content["success"]: raise Exception(content["errorMessage"])
            logger.info(f"External command executed successfully")
            
            return True
        except Exception as e:
            raise e
        
    async def execute_external_query(self, conn_key:str, query:dict | str) :
        """
        Execute an external query in the external data connection
        params:
            conn_key: str -> connection key
            query: dict | str -> query to execute
        """    
        async with SEM:
            if isinstance(query, dict): query = json.dumps(query, cls=CustomJSONEncoder)
            
            headers = {
                "Authorization": f"A2G {self.token}",
                "conn-key": conn_key,
                'Content-Type': 'text/plain'
            }
            
            timeout = httpx.Timeout(connect=10.0, read=600.0, write=600.0, pool=600.0)
            retries = 0
            backoff = INITIAL_BACKOFF
            try:
                while retries < MAX_RETRIES:
                    async with httpx.AsyncClient(http2=True, verify=VERIFY_HTTPS, timeout=timeout) as client:
                        async with client.stream("POST",f"{QUERY_MANAGER}/QueryData/ExecuteExternalQuery",
                            data=json.dumps(query, cls=CustomJSONEncoder),
                            headers = headers
                        ) as response:
                                """ if response.status_code != 200:
                                    msg = ''
                                    async for chunk in response.aiter_bytes():
                                        if chunk:
                                            msg += chunk.decode("utf-8")
                                    raise Exception(f"{response.status_code} {msg}") """
                                
                                if response.status_code != 200:
                                    if response.status_code == 404: raise Exception("ExternalDataConnection not found, please check your connkey")
                                    if response.status_code == 401: raise Exception("Unauthorized")
                                    if response.status_code == 403: raise Exception("Forbidden: please check your token or access permissions in the resource")
                                    raise Exception(f"Error executing external query, {response.status_code} {response.text}")
                                
                                
                                buffer = b""  # Buffer para ensamblar datos incompletos
                                rows_count = 0
                                async for chunk in response.aiter_bytes():
                                    buffer += chunk
                                    while len(buffer) >= 4:
                                        obj_length = struct.unpack(">I", buffer[:4])[0]
                                        if len(buffer) < 4 + obj_length:
                                            break
                                        obj_data = buffer[4:4 + obj_length]
                                        buffer = buffer[4 + obj_length:]
                                        decompressed_data = gzip.decompress(obj_data)
                                        data = msgpack.unpackb(decompressed_data, object_hook=decode_datetime)
                                        rows_count += len(data)
                                        if rows_count % 20000 == 0: logger.info(f"Downloaded {rows_count} rows")
                                        
                                        await self._write_to_file(conn_key, query, data)
                                        
                    logger.info(f"Downloaded data on retry {retries}")
                    break  # Si todo va bien, salimos del bucle
                                
            except (httpx.StreamError, httpx.ConnectError, httpx.ReadTimeout) as e:
                # Manejar errores específicos de conexión y reinicio
                retries += 1
                if retries >= MAX_RETRIES:
                    raise Exception(f"Max retries exceeded: {e}")
                logger.warning(f"Error: {e}. Retrying in {backoff} seconds...")
                await asyncio.sleep(backoff)
                backoff *= BACKOFF_MULTIPLIER  # Incrementar el tiempo de espera exponencialmente
            

    def get_external_data_conn_by_connkey(self, connkey:str) -> ExternalDataConnectionDTO:
        """
        Get external data connection by connkey
        params:
            connkey: str -> connection key
        """
        try:
            headers = { "Authorization": f"A2G {self.token}"}
            res = requests.get(APIGW + f"/ExternalDataConn/ConnKey/{connkey}", headers=headers, verify=VERIFY_HTTPS)
            
            if res.status_code != 200:
                if res.status_code == 404: raise Exception("ExternalDataConnection not found, please check your connkey")
                if res.status_code == 401: raise Exception("Unauthorized")
                if res.status_code == 403: raise Exception("Forbidden: please check your token or access permissions")
                raise Exception(f"Error getting external data connection, {res.status_code} {res.text}")
            
            content = res.json(cls=CustomJsonDecoder)
            if not content["success"]: raise Exception(content["errorMessage"])
            return ExternalDataConnectionDTO(from_response=True, **content["data"])
        except Exception as e:
            raise e

    """
    Get inputstream by ikey
    params:
        ikey: str -> inputstream key
    """
    def get_inputstream_by_ikey(self, ikey:str) -> Inputstream:
        try:
            headers = { "Authorization": f"A2G {self.token}"}

            res = requests.get(APIGW + f"/Inputstream/Ikey/{ikey}", headers=headers, verify=VERIFY_HTTPS)
            if res.status_code != 200:
                if res.status_code == 404: raise Exception("Inputstream not found, please check your ikey")
                if res.status_code == 401: raise Exception("Unauthorized")
                if res.status_code == 403: raise Exception("Forbidden: please check your token or access permissions")
                raise Exception(f"Error getting inputstream, {res.status_code} {res.text}")
            content = res.json(cls=CustomJsonDecoder)
            if not content["success"]: raise Exception(content["errorMessage"])
            return Inputstream(from_response=True, **content["data"])
        except Exception as e:
            raise e
        

    async def _write_to_file(self, ikey:str, query, data:list[dict]) -> None:
        if not os.path.exists(f".acelerai_cache/")               : os.mkdir(f".acelerai_cache/")
        if not os.path.exists(f".acelerai_cache/data/")          : os.mkdir(f".acelerai_cache/data/")
        if not os.path.exists(f".acelerai_cache/data/{ikey}/")   : os.mkdir(f".acelerai_cache/data/{ikey}/")
        
        query_str = json.dumps(query, cls=CustomJSONEncoder)
        query_hash = hashlib.sha256(query_str.encode()).hexdigest()

        # save data
        file_name = f".acelerai_cache/data/{ikey}/{query_hash}.msgpack"

        # Garantiza que solo una tarea escriba a la vez
        async with self.lock:  
            with open(file_name, "ab") as file:
                for record in data:
                    file.write(packer.pack(record))

    async def fetch_bigpage(self, ikey: str, query: dict):
        async with SEM:
            headers = {
                "Authorization": f"A2G {self.token}",
                "ikey": ikey,
                'Content-Type': 'application/json'
            }

            my_body = { "query": json.dumps(query, cls=CustomJSONEncoder) }

            buffer = b""  # Buffer para ensamblar datos incompletos
            timeout = httpx.Timeout(10800.0, connect=10.0, read=10800.0)
            retries = 0
            backoff = INITIAL_BACKOFF

            while retries < MAX_RETRIES:
                try:
                    logger.info(f"Downloading data on retry {retries}")
                    async with httpx.AsyncClient(http2=True, verify=VERIFY_HTTPS, timeout=timeout) as client:
                        async with client.stream(
                            "POST",
                            f"{QUERY_MANAGER}/QueryData/Find",
                            json=my_body,
                            headers=headers,
                            params={"page": 0, "page_size": 0}
                        ) as response:

                            if response.status_code != 200:
                                msg = ''
                                async for chunk in response.aiter_bytes():
                                    if chunk:
                                        msg += chunk.decode("utf-8")
                                raise Exception(f"{response.status_code} {msg}")

                            rows_count = 0
                            async for chunk in response.aiter_bytes():
                                buffer += chunk  # Agregar los datos al buffer
                                while len(buffer) >= 4:  # Asegurarse de que al menos 4 bytes están disponibles
                                    obj_length = struct.unpack(">I", buffer[:4])[0]

                                    if len(buffer) < 4 + obj_length:
                                        break  # Esperar más datos si el objeto no está completo

                                    obj_data = buffer[4:4 + obj_length]
                                    buffer = buffer[4 + obj_length:]  # Actualizar el buffer
                                    decompressed_data = gzip.decompress(obj_data)  # Descomprimir los datos
                                    data = msgpack.unpackb(decompressed_data, object_hook=decode_datetime)  # Deserializar el objeto
                                    rows_count += len(data)
                                    if rows_count % 20000 == 0: logger.info(f"Downloaded {rows_count} rows")

                                    # Procesar y guardar los datos
                                    await self._write_to_file(ikey, query, data)
                    logger.info(f"Downloaded data on retry {retries}")
                    break  # Si todo va bien, salimos del bucle
                except (httpx.StreamError, httpx.ConnectError, httpx.ReadTimeout) as e:
                    # Manejar errores específicos de conexión y reinicio
                    retries += 1
                    if retries >= MAX_RETRIES:
                        raise Exception(f"Max retries exceeded: {e}")
                    logger.warning(f"Error: {e}. Retrying in {backoff} seconds...")
                    await asyncio.sleep(backoff)
                    backoff *= BACKOFF_MULTIPLIER  # Incrementar el tiempo de espera exponencialmente

    async def find_one(self, ikey:str, query:dict) -> bool:
        try:
            headers = {
                "Authorization": f"A2G {self.token}",
                "ikey": ikey,
                'Content-Type': 'application/json'
            }

            async with httpx.AsyncClient(verify=VERIFY_HTTPS) as client:
                res = await client.post(f"{QUERY_MANAGER}/QueryData/FindOne", 
                    data=json.dumps(query, cls=CustomJSONEncoder),
                    headers=headers
                )

            if res.status_code != 200:
                raise Exception(f"Error getting inputstream data {res.status_code} {res.content}")
            content = res.json(cls=CustomJsonDecoder)
            await self._write_to_file(ikey, query, [content["data"]])
            return True
            
        except Exception as e:
            raise False


    async def fetch_page_agg(self, client: httpx.AsyncClient, ikey: str, pipeline:list[dict], page, page_size):
        
        logger.info(f"Downloading page {page}")
        async with SEM:
            res = await client.post(
                f"{QUERY_MANAGER}/QueryData/Aggregate",
                params = {"page": page, "page_size": page_size},
                data = json.dumps(pipeline, cls=CustomJSONEncoder)
            )

            logger.info(f"Code Res: {res.status_code}")
            if res.status_code != 200: raise Exception(f"Error getting inputstream data {res.status_code} {res.content}")
            content = res.json(cls=CustomJsonDecoder)
            
            await self._write_to_file(ikey, pipeline, content["data"])
            logger.info(f"Downloaded page {page}")


    async def aggregate(self, ikey:str, pipeline:list[dict]) -> bool:
        try:
            headers = {
                "Authorization": f"A2G {self.token}",
                "ikey": ikey,
                'Content-Type': 'application/json'
            }

            res = requests.post(f"{QUERY_MANAGER}/QueryData/ExecutionPlanningAggregate", 
                data    = json.dumps(pipeline, cls=CustomJSONEncoder), 
                headers = headers, 
                verify  = VERIFY_HTTPS
            )
            if res.status_code != 200: 
                raise Exception(f"Error getting execution planning {res.status_code} {res.content}")
            
            content = res.json(cls=CustomJsonDecoder)
            if not content["success"]: raise Exception(content["errorMessage"])
            
            total_docs      = content["data"]["total"]
            page_size       = content["data"]["size"]
            total_batchs    = (total_docs // page_size) + 1
            logger.info(f"Total documents to download {total_docs}, batch size {page_size}, total batchs {total_batchs}")
            if total_docs == 0: return False

            client = httpx.AsyncClient(headers=headers, verify=VERIFY_HTTPS, timeout=httpx.Timeout(60.0, connect=10.0, read=1200.0))
            tasks = []
            for page in range(1, total_batchs + 1):
                tasks.append(self.fetch_page_agg(client, ikey, pipeline, page, page_size))
            await asyncio.gather(*tasks)

            logger.info("All pages downloaded successfully")
            return True
            
        except Exception as e:
            logger.error(f"Error to aggregate data in inputstream {e}", stack_info=True)
            return False


    def send_example(self, ikey:str, example:dict) -> dict:
        try:
            headers = { 
                "Authorization": f"A2G {self.token}", 
                "ikey": ikey,
                'Content-Type': 'application/json'
            }
            res = requests.post(f"{DATA_URL}/Data/Insert", 
                data=json.dumps(example, cls=CustomJSONEncoder),
                headers=headers, 
                verify=VERIFY_HTTPS
            )
            if res.status_code != 200: raise Exception(f"Error sending example {res.status_code} {res.content}")
            content = res.json(cls=CustomJsonDecoder)
            if not content["success"]: raise Exception(content["errorMessage"])
            return content["data"]
        except Exception as e:
            raise e
        

    async def insert_data(self, 
        ikey:str, 
        data:list[dict], 
        mode:INSERTION_MODE, 
        wait_response:bool,
        batch_idx:int,
        batch_size:int
    ) -> tuple[int, str]:
        """
        insert data into inputstream internal
        params:
            ikey: str
            data: list[dict]
            mode: INSERTION_MODE
            wait_response: bool
            i: int -> only for logging
            batch_size: int -> only for logging
        """
        try:
            async with SEM:
                timeout = httpx.Timeout(60.0, connect=10.0, read=600.0)
                async with httpx.AsyncClient(verify=VERIFY_HTTPS, timeout=timeout) as client:

                    # TODO: revisar si se puede gzipper el contenido
                    headers = {
                        "Authorization": f"A2G {self.token}",
                        "ikey": ikey,
                        'Content-Type': 'application/json',
                        "Content-Encoding": "gzip"
                    }

                    if mode == INSERTION_MODE.REPLACE: 
                        headers["Replace"] = "true"
                        headers["Transaction"] = "false"

                    elif mode == INSERTION_MODE.INSERT_UNORDERED: 
                        headers["Replace"] = "false"
                        headers["Transaction"] = "false"

                    elif mode == INSERTION_MODE.TRANSACTION:
                        headers["Replace"] = "false"
                        headers["Transaction"] = "true"

                    if wait_response: headers["WaitResponse"] = "true"

                    # Envia un lote de datos al servidor
                    body = gzip.compress(json.dumps(data, cls=CustomJSONEncoder).encode(), compresslevel=7)
                    start = datetime.datetime.now()
                    logger.info(f"Inserting batch {batch_idx} into inputstream {ikey}...")
                    response:Response = await client.post(f"{DATA_URL}/Data/Insert", content = body, headers = headers)
                    if response.status_code != 200: 
                        raise Exception(f"Error to insert data in inputstream {response.status_code} {response.text}")
                    logger.info(f"Batch {batch_idx} received successfully, batch size: {batch_size}, elapsed time: {datetime.datetime.now() - start}")
                    return response.status_code, response.text
                
        except Exception as e:
            logger.error(f"Error to insert data in inputstream {e}", exc_info=True)
            raise e


    def convert_types(self, value):
        try:
            if value is None:
                return None
            elif isinstance(value, (int, float, str, bool)):
                if isinstance(value, float) and math.isnan(value):
                    return None  # Serializar NaN como NULL
                return value
            elif isinstance(value, Decimal):
                if value.is_nan():
                    return None  # Serializar NaN como NULL
                return str(value)
            elif isinstance(value, datetime.datetime) or isinstance(value, datetime.date):
                return value.isoformat()
            elif isinstance(value, UUID):
                return str(value)
            elif isinstance(value, bytes):
                return base64.b64encode(value).decode("utf-8")
            else:
                raise Exception(f"Tipo no soportado: {type(value)}")
        except Exception as e:
            raise(e)

    async def data_generator(self, data_source, batch_size, table_name):
        """
        Un generador asíncrono que envía datos en lotes como chunks.
        """
        batch = []
        self.sent_records = 0
        total_records = len(data_source)
        for record in data_source:
            batch.append(record)
            if len(batch) == batch_size:                            
                payload = {
                    "list_data": [{key: self.convert_types(value) for key, value in row.items()} for row in batch],
                    "table_name": table_name,
                }
                logger.info(f"Current status: {self.sent_records} of {total_records} >> sending {len(batch)}")
                self.sent_records += len(batch)
                yield msgpack.packb(payload) + b'>I<I'
                batch = []
        # Procesar los datos restantes si no alcanzaron el tamaño del batch
        if batch:                    
            payload = {
                "list_data": [{key: self.convert_types(value) for key, value in row.items()} for row in batch],
                "table_name": table_name,
            }
            logger.info(f"Current status: {self.sent_records} of {total_records} >> sending {len(batch)}")
            self.sent_records += len(batch)
            yield msgpack.packb(payload) + b'>I<I'

    async def insert_data_big_external(self, conn_key:str, table: str, data:list[dict]) -> None:
        """
        insert big data into external connection using http2 and stream
        params:
            conn_key: str -> connection key
            table_name: str -> table name
            data: list[dict] -> data to insert
        """
        batch_size = 50000
        
        try:            
            timeout = httpx.Timeout(connect=10.0, read=600.0, write=600.0, pool=600.0)
            async with httpx.AsyncClient(http2=True, verify=VERIFY_HTTPS, timeout=timeout) as client:
                # TODO: revisar si se puede gzipper el contenido
                headers = {
                    "Authorization": f"A2G {self.token}",
                    "ikey": conn_key,
                    'Content-Type': 'text/plain'
                }
                logger.info(f"Inserting data into external connection {conn_key}...")
                async with client.stream("POST",f"{QUERY_MANAGER}/QueryData/InsertDataBigExternal",
                    content = self.data_generator(data, batch_size=batch_size, table_name=table), 
                    headers = headers
                ) as response:              
                    if response.status_code == 200:
                        async for line in response.aiter_text():  # Procesar línea por línea
                            logger.info(f"Server response: {line}")
                            res_json = json.loads(line)                            
                            return res_json['success']
                    if response.status_code == 403:
                        logger.error(f"Forbidden: please check your token or access permissions in the resource") 
                        return False   
                    if response.status_code == 500:
                        logger.error(f"Internal server error: {response.text}")
                        return False
                    if response.status_code == 404:
                        logger.error(f"External data connection not found, please check your connection key")
                        return False
                    else:
                        logger.error(f"Failed to insert data: {response.status_code} - {response.text}")
                        return False
        except asyncio.CancelledError as e:
            logger.info(f"La tarea fue cancelada: {e}")
            return False
        
        except Exception as e:
            logger.info(f"Error general: {e}")     
            logger.info(f"Error general: {repr(e), {traceback.format_exc()}}")  
            return False
        return True


    async def remove_documents(self, ikey: str, query: dict) -> int:
        try:
            logger.info("Removing data...")
            headers = {
                "Authorization": f"A2G {self.token}",
                "ikey": ikey,
                'Content-Type': 'application/json'
            }

            if len(query) == 0: 
                raise Exception("Query is empty, please provide a valid query. If you desire to delete all documents, use the clear_inputstream method.")

            async with httpx.AsyncClient(verify=VERIFY_HTTPS) as client:
                response = await client.post(
                    f"{QUERY_MANAGER}/QueryData/RemoveDocuments",
                    data=json.dumps(query, cls=CustomJSONEncoder),
                    headers=headers,
                    timeout = httpx.Timeout(60.0, connect=10.0, read=600.0)
                )

            if response.status_code != 200: 
                raise Exception(f"Error to remove data in inputstream {response.status_code} {response.content}")
            
            res_object = json.loads(response.text, cls=CustomJsonDecoder)
            if not res_object["success"]: 
                raise Exception(res_object["errorMessage"])

            content = res_object["data"]
            deleted_docs = content["docs_affected"]
            logger.info(f"Operation complete, total docs deleted: {deleted_docs}")

            return deleted_docs
        except Exception as e:
            raise e


    async def clear_inputstream(self, ikey:str) -> int:
        try:
            logger.info("Removing all data...")
            headers = {
                "Authorization": f"A2G {self.token}",
                "ikey": ikey
            }

            async with httpx.AsyncClient(verify=VERIFY_HTTPS) as client:
                response = await client.post(
                    f"{QUERY_MANAGER}/QueryData/Clear",
                    headers=headers,
                    timeout = httpx.Timeout(60.0, connect=10.0, read=600.0)
                )

            if response.status_code != 200:
                raise Exception(f"Error to remove all data in inputstream {response.status_code} {response.content}")
            
            res_object = response.json(cls=CustomJsonDecoder)
            if not res_object["success"]: raise Exception(res_object["errorMessage"])
            
            content = res_object["data"]
            deleted_docs = content["docs_affected"]
            logger.info(f"Operation complete, total docs deleted: {deleted_docs}")
            return deleted_docs
        except Exception as e:
            logger.error(f"Error to clear inputstream {e}", exc_info=True)
            raise e
        


    ##############################################################
    
    def upload_asset(self, pkey:str, local_file_path:str, storage_path:str) -> Asset:
        """
        Upload file to the asset storage
        Please note: if your connection is to MongoDB, the query must be as follows: {"collection_name":"your_collection", "query":'{your_query}'}
        params:
            conn_key: str
            query: dict | str
            cache: bool = True -> if True, use cache if exists and is not expired
        """
        headers = { "Authorization": f"A2G {self.token}"}
        
        file_stream = open(local_file_path, 'rb')
        files = {'file': (f'{local_file_path.split("/")[-1]}', file_stream)}
        
        endpoint = f"{APIGW}/Project/{pkey}/Assets/Upload"
        response = requests.post(endpoint, data = {'path': storage_path}, files = files, headers=headers, verify=VERIFY_HTTPS)
        content = response.json(cls=CustomJsonDecoder)
        if response.status_code == 200:
            logger.info(f"Asset replaced successfully")
            if not content["success"]: raise Exception(content["errorMessage"])
            asset_data = content["data"]
            return Asset(
                id=asset_data["id"],
                path=asset_data["path"],
                created_at=asset_data["createdAt"],
                url=asset_data["url"],
                size=asset_data["size"]
            )
        elif response.status_code == 201:
            logger.info(f"Asset created successfully")
            if not content["success"]: raise Exception(content["errorMessage"])
            asset_data = content["data"]
            return Asset(
                id=asset_data["id"],
                path=asset_data["path"],
                created_at=asset_data["createdAt"],
                url=asset_data["url"],
                size=asset_data["size"]
            )
        else:
            logger.error(f"Request failed with status code {response.status_code}")
            raise Exception(f"Error uploading asset {response.status_code} {response.text}")


    def get_asset_by_id(self, uid:str) -> Asset:
        """
        Get asset by its unique id
        params:
            uid: str -> unique identifier of the asset
        """
        headers = {"Authorization": f"A2G {self.token}"}
        endpoint = f"{APIGW}/Assets/{uid}"
        response = requests.get(endpoint, headers=headers, verify=VERIFY_HTTPS)
        if response.status_code != 200:
            logger.error(f"Request failed with status code {response.status_code}")
            raise Exception(f"Error getting asset {response.status_code} {response.text}")
        content = response.json(cls=CustomJsonDecoder)
        if not content["success"]:
            raise Exception(content["errorMessage"])
        asset_data = content["data"]
        return Asset(
            id=asset_data["id"],
            path=asset_data["path"],
            created_at=asset_data["createdAt"],
            url=asset_data["url"],
            size=asset_data["size"]
        )


    def get_asset_by_path(self, pkey:str, path:str) -> Asset:
        """
        Get asset by project key and path
        params:
            pkey: str -> project key
            path: str -> path of the asset
        """
        headers = {"Authorization": f"A2G {self.token}"}
        endpoint = f"{APIGW}/Project/{pkey}/Assets/{path}"
        response = requests.get(endpoint, headers=headers, verify=VERIFY_HTTPS)
        if response.status_code != 200:
            logger.error(f"Request failed with status code {response.status_code}")
            raise Exception(f"Error getting asset by path {response.status_code} {response.text}")
        content = response.json(cls=CustomJsonDecoder)
        if not content["success"]:
            raise Exception(content["errorMessage"])
        asset_data = content["data"]
        return Asset(
            id=asset_data["id"],
            path=asset_data["path"],
            created_at=asset_data["createdAt"],
            url=asset_data["url"],
            size=asset_data["size"]
        )


    def download_userfile(self, url:str, save_path:str):
        """
        Download a file from a given URL and save it to the specified path using streaming.
        :param url: The URL of the file to download.
        :param save_path: The local path where the file will be saved.
        """
        headers = {"Authorization": f"A2G {self.token}"}
        with requests.get(url, headers=headers, verify=VERIFY_HTTPS, stream=True) as r:
            r.raise_for_status()
            with open(save_path, "wb") as f:
                for chunk in r.iter_content(chunk_size = 1024*1024):  # 1 MB chunks
                    if chunk:
                        f.write(chunk)
        logger.info(f"File downloaded successfully to {save_path}")
    

    def list_assets(self, pkey:str, page:int=1, page_size:int=100, folder_path:str = None, text_search:str = None) -> PaginationData[Asset]:
        """
        List assets in a project with pagination and optional filters.
        params:
            pkey: str -> project key
            page: int -> page number (default is 1)
            page_size: int -> number of items per page (default is 100)
            folder_path: str -> optional folder path to filter assets
            text_search: str -> optional text search to filter assets
        """
        headers = { "Authorization": f"A2G {self.token}", }

        params = { "page": page, "pageSize": page_size }
        if folder_path is not None: params["folderPath"] = folder_path
        if text_search is not None: params["textSearch"] = text_search
        endpoint = f"{APIGW}/Project/{pkey}/Assets"
        response = requests.get(endpoint, headers=headers, verify=VERIFY_HTTPS, params=params)
        api_response = response.json(cls=CustomJsonDecoder)

        if response.status_code == 200:
            assets_list = api_response['data']['data']
            total = api_response['data']['total']

            assets:list[Asset] = []
            for asset in assets_list:
                asset = Asset(
                    id=asset['id'],
                    path=asset['path'],
                    created_at=asset['createdAt'],
                    url=asset['url'],
                    size=asset['size']
                )
                assets.append(asset)

            page_data = PaginationData(
                data=assets,
                total_items=total,
                page=page,
                page_size=page_size
            )

            return page_data
        else:
            logger.error(f"Request failed with status code {response.status_code}")

        return response.json()


    def delete_asset_by_id(self, asset_id:str) -> bool:
        """
        Delete an asset by its unique id.
        params:
            asset_id: str -> unique identifier of the asset
        """
        headers = {"Authorization": f"A2G {self.token}"}
        endpoint = f"{APIGW}/Assets/{asset_id}"
        response = requests.delete(endpoint, headers=headers, verify=VERIFY_HTTPS)
        
        if response.status_code == 200:
            logger.info(f"Asset {asset_id} deleted successfully.")
            return True
        else:
            logger.error(f"Failed to delete asset {asset_id}: {response.status_code} {response.text}")
            return False
        

    def delete_asset_by_path(self, pkey:str, path:str) -> bool:
        """
        Delete an asset by project key and path.
        params:
            pkey: str -> project key
            path: str -> path of the asset
        """
        headers = {"Authorization": f"A2G {self.token}"}
        endpoint = f"{APIGW}/Project/{pkey}/Assets/{path}"
        response = requests.delete(endpoint, headers=headers, verify=VERIFY_HTTPS)
        content = response.json(cls=CustomJsonDecoder)
        if response.status_code == 200:
            logger.info(f"Asset at path {path} in project {pkey} deleted successfully.")
            return True
        else:
            logger.error(f"Failed to delete asset at path {path} in project {pkey}: {response.status_code} {response.text}")
            return False


    def download_folder(self, pkey:str, folder_path:str, save_path:str):
        """
        Download all assets in a folder to a local path.
        params:
            pkey: str -> project key
            folder_path: str -> path of the folder in the project
            local_path: str -> local path where the assets will be downloaded
        """
        
        # get folders
        dirs = save_path.split('/')
        if len(dirs) > 1:
            local_folder_path = '/'.join(dirs[:-1])
            if not os.path.exists(local_folder_path):
                os.makedirs(local_folder_path)

        headers = {"Authorization": f"A2G {self.token}"}
        endpoint = f"{APIGW}/Project/{pkey}/Assets/Folder/{folder_path}"
        with requests.get(endpoint, headers=headers, verify=VERIFY_HTTPS, stream=True) as r:
            r.raise_for_status()
            with open(save_path, "wb") as f:
                for chunk in r.iter_content(chunk_size = 1024*1024):  # 1 MB chunks
                    if chunk:
                        f.write(chunk)
        logger.info(f"File downloaded successfully to {save_path}")


import copy
from datetime import datetime
from enum import Enum
from uuid import UUID

# Enums
class FileIndexFieldType(Enum):
    Datetime = 0
    String = 1
    Number = 2
    Integer = 3

class DateBucketSize(Enum):
    Minute = 0
    Hour = 1
    Day = 2
    Week = 3
    Month = 4
    Year = 5

class InputstreamStatus(Enum):
    ToDiscover = 0
    Undiscovered = 1
    Exposed = 2
    ToDiscoverAgain = 3

class InputstreamStorage(Enum):
    Collection = 0
    TimeSeriesCollection = 1
    File = 2

class InputstreamProtocol(Enum):
    MQTT = 0
    HTTP = 1
    BOTH = 2

class RealTimeMode(Enum):
    OFF = 0
    ON = 1

class IndexType(Enum):
    Unique = 0
    Search = 1

class SortType(Enum):
    Ascending = 0
    Descending = 1

class SourceType(Enum):
    MySQL=0
    MongoDB=1
    SQLServer=2
    Snowflake=3
    Oracle=4
    PostgresSQL=5
    Firebase=6
    BigQuery=7

class DataType(Enum):
    TypeNumber = 0
    TypeBoolean = 1
    TypeString = 2
    TypeDateType = 3
    TypeDate = 4
    TypeList = 5

class InputstreamType(Enum):
    InSystem = 0
    Native = 1

# Models
class DynamicField:
    DataType: DataType
    Field: str
    ValueDefault: str

    def __init__(self, **kwargs) -> None:
        kwargs["DataType"] = DataType(kwargs.pop('DataType'))
        self.__dict__ = kwargs

    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        return data 

class DataConnection:
    ConnectionString: str
    DataSourceName: str
    Query : str
    SourceType: SourceType
    DynamicFields: list[DynamicField]
    DatabaseName: str

    def __init__(self,**kwargs)-> None:
        kwargs["SourceType"] = SourceType(kwargs.pop('SourceType'))
        self.__dict__ = kwargs
    
    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        return data


class IndexField:
    Name: str
    FieldType: FileIndexFieldType
    DoubleBucketSize: float
    DateBucketSize: DateBucketSize

    def __init__(self, from_response=False, **kwargs):
        if from_response:
            kwargs["Name"] = kwargs.pop('name')
            kwargs["DoubleBucketSize"] = kwargs.pop('doubleBucketSize')
            kwargs["FieldType"] = FileIndexFieldType(kwargs.pop('fieldType'))
            kwargs["DateBucketSize"] = DateBucketSize(kwargs.pop('dateBucketSize'))
        else:
            kwargs["FieldType"] = FileIndexFieldType(kwargs.pop('FieldType'))
            kwargs["DateBucketSize"] = DateBucketSize(kwargs.pop('DateBucketSize'))
        self.__dict__ = kwargs

    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        return data


class CollectionIndexField:
    Name: str
    SortType: SortType

    def __init__(self, from_response=False, **kwargs):
        if from_response:
            kwargs["Name"] = kwargs.pop('name')
            kwargs["SortType"] = SortType(kwargs.pop('sortType'))
        else:
            kwargs["SortType"] = SortType(kwargs["SortType"])
        self.__dict__ = kwargs

    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        return data


class CollectionIndex:
    Name: str
    Fields: list[CollectionIndexField]
    Size: int
    IndexUse: int
    SinceUse: datetime
    IndexType: IndexType
    DateCreated: datetime
    IsCompound: bool

    def __init__(self, from_response=False, **kwargs):
        if from_response:
            kwargs["Name"] = kwargs.pop('name')
            kwargs["DateCreated"] = kwargs.pop('dateCreated')
            kwargs["Fields"] = [CollectionIndexField(from_response=True, **x) for x in kwargs.pop('fields')]
            kwargs["IndexType"] = IndexType(kwargs.pop('indexType'))
        else:
            kwargs["DateCreated"] = kwargs.pop('DateCreated')
            kwargs["Fields"] = [CollectionIndexField(**x) for x in kwargs["Fields"]]
            kwargs["IndexType"] = IndexType(kwargs.pop('IndexType'))
        self.__dict__ = kwargs

    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        data["Fields"] = [x.get_dict() for x in self.Fields]
        return data
    
class ExternalDataConnectionDTO:
    Id: UUID 
    ConnectionKey: str 
    Name: str 
    DataSourceName: str 
    DatabaseName: str 
    Tags: list[str] 
    DatabaseType: SourceType 
    CreatedOn: datetime 
    Removed: bool 
    
    def __init__(self, from_response = False, **kwargs):
        if from_response: self.from_response(**kwargs)
        else:
            kwargs["Id"] = UUID(kwargs.pop('Id'))
            kwargs["ConnectionKey"] = kwargs.pop('ConnectionKey')
            kwargs["Name"] = kwargs.pop('Name')
            kwargs["DataSourceName"] = kwargs.pop('DataSourceName')
            kwargs["DatabaseName"] = kwargs.pop('DatabaseName')
            kwargs["Tags"] = kwargs.pop('Tags')
            kwargs["DatabaseType"] = SourceType(kwargs.pop('DatabaseType'))
            kwargs["CreatedOn"] = kwargs.pop('CreatedOn')

            self.__dict__ = kwargs
            
    def from_response(self, **kwargs) -> None:
        kwargs["Id"] = UUID(kwargs.pop('id'))
        kwargs["ConnectionKey"] = kwargs.pop('connectionKey')
        kwargs["Name"] = kwargs.pop('name')
        kwargs["DatabaseName"] = kwargs.pop('databaseName')
        kwargs["DatabaseType"] = SourceType(kwargs.pop('databaseType'))
        kwargs["CreatedOn"] = kwargs.pop('createdOn')

        self.__dict__ = kwargs
        
    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        data["Id"] = str(data["Id"])
        return data

class Inputstream:
    Id: UUID
    SubscriptionId: UUID
    Name: str
    CollectionName: str
    CollectionMongo: str | None
    Schema: str
    SchemaSample: str
    SampleDate: datetime
    Status: InputstreamStatus
    InputstreamType: InputstreamType
    DataConnection: DataConnection
    ExternalDataConnId: UUID
    Tags: list[str]
    Ikey: str
    CollectionIndexes: list[CollectionIndex]
    FilesIndex: list[IndexField]
    Storage: InputstreamStorage
    Protocol: InputstreamProtocol
    RealTimeMode: RealTimeMode
    Size: int
    MaxNDocsByFile: int
    AllowAnyOrigin: bool
    FileConsolidatorCron: str
    Removed: bool
    CreatedOn: datetime
    RemovedOn: datetime | None


    def __init__(self, from_response = False, **kwargs):
        if from_response: self.from_response(**kwargs)
        else:
            kwargs["Id"] = UUID(kwargs.pop('Id'))
            """ kwargs["ExternalDataConnId"] = UUID(kwargs.pop('ExternalDataConnId')) """
            kwargs["ExternalDataConnId"] = UUID(kwargs.pop('ExternalDataConnId')) if kwargs.get('ExternalDataConnId') else None
            
            kwargs["Status"]       = InputstreamStatus(kwargs.pop('Status'))
            kwargs["DataConnection"] = DataConnection(**kwargs["DataConnection"])
            kwargs["InputstreamType"] = InputstreamType(kwargs.pop('InputstreamType'))
            kwargs["Storage"]      = InputstreamStorage(kwargs.pop('Storage'))
            kwargs["Protocol"]     = InputstreamProtocol(kwargs.pop('Protocol'))
            kwargs["RealTimeMode"] = RealTimeMode(kwargs.pop('RealTimeMode'))

            kwargs["FilesIndex"]        = [IndexField(**x) for x in kwargs["FilesIndex"]]
            kwargs["CollectionIndexes"] = [CollectionIndex(**x) for x in kwargs["CollectionIndexes"]]

            self.__dict__ = kwargs

    
    def from_response(self, **kwargs) -> None:
        kwargs["Id"]           = UUID(kwargs.pop('id'))
        kwargs["SubscriptionId"]            = UUID(kwargs.pop('subscriptionId'))
        kwargs["ExternalDataConnId"]        = UUID(kwargs.pop('externalDataConnId')) if kwargs.get('externalDataConnId') else None

        kwargs["Name"]                      = kwargs.pop('name')
        kwargs["CollectionName"]            = kwargs.pop('collectionName')
        kwargs["CollectionMongo"]           = kwargs.pop('collectionMongo')
        kwargs["Schema"]                    = kwargs.pop('schema')
        kwargs["SchemaSample"]              = kwargs.pop('schemaSample')
        kwargs["Tags"]                      = kwargs.pop('tags')
        kwargs["Ikey"]                      = kwargs.pop('ikey')
        kwargs["Size"]                      = kwargs.pop('size')
        kwargs["MaxNDocsByFile"]            = kwargs.pop('maxNDocsByFile')
        kwargs["AllowAnyOrigin"]            = kwargs.pop('allowAnyOrigin')
        kwargs["FileConsolidatorCron"]      = kwargs.pop('fileConsolidatorCron')
        kwargs["Removed"]                   = kwargs.pop('removed')

        kwargs["Status"]       = InputstreamStatus(kwargs.pop('status'))
        kwargs["Storage"]      = InputstreamStorage(kwargs.pop('storage'))
        kwargs["Protocol"]     = InputstreamProtocol(kwargs.pop('protocol'))
        kwargs["RealTimeMode"] = RealTimeMode(kwargs.pop('realTimeMode'))
        kwargs["InputstreamType"] = InputstreamType(kwargs.pop('inputstreamType'))

        kwargs["FilesIndex"]        = [IndexField(from_response=True,**x) for x in kwargs.pop("filesIndex")]
        kwargs["CollectionIndexes"] = [CollectionIndex(from_response=True, **x) for x in kwargs.pop("collectionIndexes")]

        kwargs["SampleDate"]  = kwargs.pop('sampleDate')
        kwargs["CreatedOn"]   = kwargs.pop('createdOn') 
        kwargs["RemovedOn"]   = kwargs.pop('removedOn') 

        self.__dict__ = kwargs

    def get_dict(self):
        data = copy.deepcopy(self.__dict__)
        data["Id"] = str(data["Id"])
        data["FilesIndex"] = [x.get_dict() for x in self.FilesIndex]
        data["CollectionIndexes"] = [x.get_dict() for x in self.CollectionIndexes]
        return data


class INSERTION_MODE(Enum):
    """
    Enum for insertion modes, available modes: 
    REPLACE: if a document collides with an existing document by a unique index, the existing document is replaced with the new document. Otherwise, the new document is inserted.
    INSERT_UNORDERED: insert all documents that did not have a collision with an existing document by a unique index.
    TRANSACTION: insert all documents in a transaction, if a document collides with an existing document by a unique index, the transaction is aborted and no document is inserted.
    """
    REPLACE = 0
    INSERT_UNORDERED = 1
    TRANSACTION = 2




class Asset:
    """
    Represents an asset in the AcelerAI.
    Attributes:
        id (str): Unique identifier for the asset.
        path (str): path of the asset.
        created_on (str): Creation date of the asset.
    """
    id: str
    path: str
    created_at: str
    url: str
    size: int

    def __init__(self, id: str, path: str, created_at: str, url: str, size:int):
        """
        Initializes the Asset with id, path, and creation date.
        :param id: Unique identifier for the asset.
        :param path: path of the asset.
        :param created_at: Creation date of the asset.
        """
        self.id = id
        self.url = url
        self.path = path
        self.size = size
        self.created_at = created_at


import json
import os

__mode = os.environ.get("EXEC_LOCATION", "LOCAL")
if __mode == "LOCAL":
    # Environment production
    DATA_URL        = os.environ.get("DATA_URL"         , "https://stream.aceler.ai")
    QUERY_MANAGER   = os.environ.get("QUERY_MANAGER"    , "https://stream.aceler.ai")  
    APIGW           = os.environ.get("APIGW"            , "https://apigw.aceler.ai")
    VERIFY_HTTPS    = os.environ.get("VERIFY_HTTPS", "true").lower().strip() == "true"

elif __mode == "PRODUCTION":
    # Environment production
    DATA_URL        = os.environ.get("DATA_URL"         , "https://inputstreamdata-service.aceler-ai.svc.cluster.local:1008")
    QUERY_MANAGER   = os.environ.get("QUERY_MANAGER"    , "https://querymanager-service.aceler-ai.svc.cluster.local:1012")
    APIGW           = os.environ.get("APIGW"            , "https://apigateway-service.aceler-ai.svc.cluster.local:1000")
    VERIFY_HTTPS    = False

else:
    raise Exception(f"Invalid environment {__mode}, please check the EXEC_LOCATION environment variable")

from acelerai_pipelines.pipelines_client import PipelinesClient
from acelerai_pipelines.models.run import Run
from acelerai_pipelines.models.deploy import Deploy

__results_path = os.environ.get("A2G_RESULT_PATH","a2g_results")
__payload_path = os.environ.get("A2G_PAYLOAD_PATH", "payload.json")

__mode = os.environ.get("EXEC_LOCATION", "LOCAL")

def save_result(key:str, value, path = None):
    """
    Save the result in the file
    :param key: The key to be used to save the result
    :param value: The value to be saved
    :param path: The path to save the result, if None, the default path is used
    """
    result_path = __results_path
    if path is not None and __mode == "LOCAL":
        result_path = path

    if __mode == "LOCAL":
        if not os.path.exists(result_path): os.makedirs(result_path)

    open(f"{result_path}/{key}", 'w+').write(json.dumps(value))

def get_payload(path = None) -> dict | None:
    """
    Get the payload from the file, if the file does not exist, return None
    :param path: The path to the payload file, if None, the default path is used
    """
    payload_path = __payload_path
    if path is not None and __mode == "LOCAL":
        payload_path = path

    if not os.path.exists(payload_path): return None
    return json.loads(open(payload_path).read())


from typing import TypeVar, Generic

T = TypeVar('T')
class PaginationData(Generic[T]):
    """
    Pagination data class
    """
    data: list[T]
    page: int
    page_size: int
    total: int

    def __init__(self, data: list[T], page: int, page_size: int, total: int):
        self.data = data
        self.page = page
        self.page_size = page_size
        self.total = total


from typing import TypeVar, Generic, Optional
T = TypeVar('T')

class PaginationData(Generic[T]):
    """
    A class to represent paginated data.

    Attributes:
        total_items (int): Total number of items across all pages.
        total_pages (int): Total number of pages.
        page (int): Current page number.
        page_size (int): Number of items per page.
        data (list[T]): List of items in the current page.
    """

    total_items: int
    total_pages: int
    page: int
    page_size: int
    data: list[T]

    def __init__(self, data: list[T], total_items:int, page:int, page_size:int):
        """
        Initializes the PaginationData with items, total count, current page, and page size.
        :param items: List of items in the current page.
        :param total: Total number of items across all pages.
        :param page: Current page number.
        :param page_size: Number of items per page.
        """
        self.data = data
        self.total_items = total_items
        self.total_pages = (total_items + page_size - 1) // page_size
        self.page = page
        self.page_size = page_size

    def __repr__(self):
        return f"total_pages={self.total_pages}, total_items={self.total_items}, page={self.page}, page_size={self.page_size}, len_page={len(self.data)}"
    
    def __len__(self):
        """
        Returns the number of items in the current page.
        """
        return len(self.data)


import os
import asyncio
from  acelerai_inputstream import InputstreamClient

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')

async def main():
    client = InputstreamClient(token='7fd3640954424cba8518309e11939027')
    test = await client.find("e5803b1c08364d7c8b37", {})
    print(len(test))
    print(test[0])

if __name__ == '__main__':
    asyncio.run(main())


from copy import deepcopy
from uuid import UUID
from enum import Enum
from datetime import datetime

from acelerai_pipelines.util import toUpperFirst
from .run_trigger_type import RunTriggerType
from .environment_item import EnvironmentItem


class DeployType(Enum):
    OnDemand = 0
    OnDemandAndScheduled = 1


class DeployInfo:
    FKey: str
    CronExpression: str
    Actived: bool
    DeployType: DeployType

    def __init__(self, **kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        kwargs["DeployType"] = DeployType(kwargs["DeployType"])
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["DeployType"] = self.DeployType.value
        return data


class DeployParams:
    NodesIds: list[UUID]
    RunImageId: UUID
    GitServerId: UUID
    ProjectGitId: UUID
    ForceWriteLogs: bool
    LastCommitSelected: bool
    CommitHash: str
    CommitComment: str
    Environments: list[EnvironmentItem]
    Notebooks: list[str]
    Timeouts: list[int]
    PayloadSchema: str

    def __init__(self, **kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        kwargs["Environments"] = [EnvironmentItem(**env) for env in kwargs["Environments"]]
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["Environments"] = [env.get_dict() for env in self.Environments]
        return data


class CommunicationInfo:
    Inputs: list
    Outputs: list
    Communications: list

    def __init__(self,**kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        return data


class StatusControl:
    MinutesInterval: int
    LastError: datetime
    LastOk: datetime
    OkRunId: UUID
    ErrorRunId: UUID
    OkTriggerType: RunTriggerType
    ErrorTriggerType: RunTriggerType

    def __init__(self, **kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        kwargs["OkTriggerType"] = RunTriggerType(kwargs["OkTriggerType"]) if kwargs["OkTriggerType"] else None
        kwargs["ErrorTriggerType"] = RunTriggerType(kwargs["ErrorTriggerType"]) if kwargs["ErrorTriggerType"] else None
        self.__dict__ = kwargs
    
    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["OkTriggerType"] = self.OkTriggerType.value if self.OkTriggerType else None
        data["ErrorTriggerType"] = self.ErrorTriggerType.value if self.ErrorTriggerType else None
        return data


class Deploy:
    Id: UUID
    SubscriptionId: UUID
    ProjectId: UUID
    Name: str
    Description: str
    DeployInfo: DeployInfo
    DeployParams: DeployParams
    CommunicationInfo: CommunicationInfo
    StatusControl: StatusControl
    CreatedOn: datetime
    RemovedOn: datetime
    Removed: bool

    def __init__(self,**kwargs):
        # first letter of to upper case
        kwargs = {toUpperFirst(k) : v for k, v in kwargs.items()}
        kwargs["DeployInfo"] = DeployInfo(**kwargs["DeployInfo"])
        kwargs["DeployParams"] = DeployParams(**kwargs["DeployParams"])
        kwargs["CommunicationInfo"] = CommunicationInfo(**kwargs["CommunicationInfo"]) if kwargs["CommunicationInfo"] else None
        kwargs["StatusControl"] = StatusControl(**kwargs["StatusControl"])

        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["DeployInfo"] = self.DeployInfo.get_dict()
        data["DeployParams"] = self.DeployParams.get_dict()
        data["CommunicationInfo"] = self.CommunicationInfo.get_dict() if self.CommunicationInfo else None
        data["StatusControl"] = self.StatusControl.get_dict()
        return data



from copy import deepcopy
from enum import Enum
from uuid import UUID
from datetime import datetime
from acelerai_pipelines.util import toUpperFirst
from .run_trigger_type import RunTriggerType
from .environment_item import EnvironmentItem


class RunStatus(Enum):
    Pending = 0
    Running = 1
    Error = 2
    Completed = 3
    Cancelled = 4
    InternalError = 5
    Lost = 6


class StepStatus(Enum):
    Waiting = 0
    Building = 1
    Running = 2
    Error = 3
    Success = 4


class RunParams:
    NodesIds: list[UUID]
    ProjectId: UUID
    ProjectName: str
    RunImageId: UUID
    GitServerId: UUID
    ProjectGitId: UUID
    LastCommitSelected: bool
    ForceWriteLogs: bool
    CommitHash: str
    CommitComment: str
    Environments: list[EnvironmentItem]
    Notebooks: list[str]
    Timeouts: list[int]
    Payload: str

    def __init__(self, **kwargs):
        kwargs = {k.capitalize(): v for k, v in kwargs.items()}
        kwargs["Environments"] = [EnvironmentItem(**env) for env in kwargs["Environments"]]
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["Environments"] = [env.get_dict() for env in self.Environments]
        return data


class StepControlProps:
    Name: str
    FilePath: str
    StartTime: datetime
    EndTime: datetime
    StepLogs: str
    StepStatus: StepStatus
    JupyterOut: str
    LogsSize: int
    HTMLSize: int

    def __init__(self,**kwargs):
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}
        kwargs["StepStatus"] = StepStatus(kwargs["StepStatus"])
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["StepStatus"] = self.StepStatus.value
        return data


class Run:
    Id: UUID
    SubscriptionId: UUID
    DeployId: UUID
    RunTriggerType: RunTriggerType
    RunStatus: RunStatus
    CreatedOn: datetime
    EnqueuedTime: datetime
    SocketRoomName: str
    RunParams: RunParams
    StepsControlProps: list[StepControlProps]
    ImageStartTime: datetime
    ImageFinishTime: datetime
    LastPing: datetime
    Result:str
    PodName: str
    NodeName: str

    def __init__(self, **kwargs):
        # capitalize all keys
        kwargs = {toUpperFirst(k): v for k, v in kwargs.items()}

        kwargs["RunTriggerType"] = RunTriggerType(kwargs["RunTriggerType"])
        kwargs["RunStatus"] = RunStatus(kwargs["RunStatus"])
        kwargs["RunParams"] = RunParams(**kwargs["RunParams"])
        kwargs["StepsControlProps"] = [StepControlProps(**step) for step in kwargs["StepsControlProps"]] if kwargs["StepsControlProps"] else None
        self.__dict__ = kwargs

    def get_dict(self):
        data = deepcopy(self.__dict__)
        data["RunTriggerType"] = self.RunTriggerType.value
        data["RunStatus"] = self.RunStatus.value
        data["RunParams"] = self.RunParams.get_dict()
        data["StepsControlProps"] = [step.get_dict() for step in self.StepsControlProps]
        return data



from datetime import date, datetime
from enum import Enum
import json
import math
from uuid import UUID
import msgpack
from decimal import Decimal
from dateutil import parser

import logging
logger = logging.getLogger("AcelerUtil")

def custom_encoder(obj):
    if isinstance(obj, datetime) or isinstance(obj, date):
        return obj.isoformat()
        
    if isinstance(obj, Decimal):
        return float(obj)
    
    raise TypeError(f"Object of type {type(obj).__name__} is not serializable")


def decode_datetime(obj:dict):
    for k, v in obj.items():
        if isinstance(v, str) and 'T' in v and '-' in v and ':' in v and len(v) < 35:
            try:
                dv = parser.parse(v)
                dt = dv.replace(tzinfo=None)
                obj[k] = dt
            except: pass

        elif isinstance(v, str) and '-' in v and len(v) < 11:
            try:
                obj[k] = parser.parse(v).date()
                if 'T' in v and '-' in v and ':' in v and len(v) < 40:
                    obj[k] = datetime.fromisoformat(v)
                elif '-' in v and len(v) < 11:
                    obj[k] = parser.parse(v)
            except ValueError:
                pass

    return obj


def load_full_object(file_path):
    """Carga completamente el objeto desde un archivo MessagePack en memoria. Cargar todos los registros en memoria como una lista de diccionarios."""
    try:
        with open(file_path, "rb") as file:
            unpacker = msgpack.Unpacker(file, raw=False, object_hook=decode_datetime)
            data = [record for record in unpacker]
        return data
    except Exception as e:
        logger.error(f"Error al cargar el archivo: {e}", exc_info=True)
        return None
    

class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, float) and math.isnan(obj):
            return None
        
        elif obj is None:
            return None
        
        elif isinstance(obj, Enum):
            return obj.value
        
        elif isinstance(obj, datetime):
            return obj.isoformat()
        
        elif isinstance(obj, date):
            #logger.debug(f"Date: {obj}")
            #return datetime.strptime(obj, "%Y-%m-%d").date()
            return obj.isoformat()
        
        elif isinstance(obj, UUID):
            return str(obj)
        else:
            return super().default(obj)


class CustomJsonDecoder(json.JSONDecoder):
    def __init__(self, *args ,**kargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kargs)

    def object_hook(self, obj:dict):
        for k, v in obj.items():
            if isinstance(v, str) and 'T' in v and '-' in v and ':' in v and len(v) < 40:
                try:
                    dv = parser.parse(v)
                    dt = dv.replace(tzinfo=None)
                    obj[k] = dt
                except:
                    pass
            elif isinstance(v, str) and '-' in v and len(v) < 11:
                try:
                    obj[k] = parser.parse(v).date()
                except:
                    pass
        return obj






import json
import os
import logging
logger = logging.getLogger(__name__)

__mode = os.environ.get("EXEC_LOCATION", "LOCAL")
logger.info(f"EXEC_LOCATION: {__mode}")
if __mode == "LOCAL":
    # Environment production
    DATA_URL        = os.environ.get("DATA_URL"         , "https://stream.aceler.ai")
    QUERY_MANAGER   = os.environ.get("QUERY_MANAGER"    , "https://stream.aceler.ai")  
    APIGW           = os.environ.get("APIGW"            , "https://apigw.aceler.ai")
    VERIFY_HTTPS    = os.environ.get("VERIFY_HTTPS", "true").lower().strip() == "true"

elif __mode == "PRODUCTION":
    # Environment production
    DATA_URL        = os.environ.get("DATA_URL"         , "https://inputstreamdata-service.aceler-ai.svc.cluster.local:1008")
    QUERY_MANAGER   = os.environ.get("QUERY_MANAGER"    , "https://querymanager-service.aceler-ai.svc.cluster.local:1012")
    APIGW           = os.environ.get("APIGW"            , "https://apigateway-service.aceler-ai.svc.cluster.local:1000")
    VERIFY_HTTPS = False

else:
    raise Exception(f"Invalid environment {__mode}, please check the EXEC_LOCATION environment variable")


from acelerai_inputstream.models.inputstream import INSERTION_MODE, Inputstream
from acelerai_inputstream.inputstream_client import InputstreamClient

__mode = os.environ.get("EXEC_LOCATION", "LOCAL")


from acelerai_inputstream import InputstreamClient
import logging

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')

#Para desactivar los logs de la librería
#logging.getLogger('InputstreamClient').disabled = True
client = InputstreamClient(token='7ffba0a8bccb4498be8a811b43bd9400') 
input_cycle=client.find(ikey='31feaa0361b44ebb8b70', query= {}, cache=False)






import asyncio
import json
import httpx
import msgpack
from acelerai_pipelines import APIGW, VERIFY_HTTPS
from acelerai_pipelines.models.deploy import Deploy
from acelerai_pipelines.models.run import Run
from acelerai_inputstream.utils import CustomJSONEncoder, CustomJsonDecoder, custom_encoder

# Type conversion

import logging

from acelerai_pipelines.models.pagination_data import PaginationData
logger = logging.getLogger("HttpAcelerAI")

SEM = asyncio.Semaphore(20)  # Limitar concurrencia a 20 conexiones
MAX_RETRIES = 5  # Número máximo de reintentos
INITIAL_BACKOFF = 1  # Tiempo inicial de espera en segundos
BACKOFF_MULTIPLIER = 2  # Factor multiplicador para el tiempo de espera

packer = msgpack.Packer(default=custom_encoder) 


class AcelerAIHttpPipelinesClient():

    def __init__(self, token:str):
        self.token = token

    
    def get_headers(self) -> dict:
        """
        Get the headers for the request
        :return: The headers
        """
        return {
            "Authorization": f"A2G {self.token}",
            "Content-Type": "application/json"
        }


    async def get_deploy_by_fkey(self, fkey:str) -> Deploy | None:
        """
        Get deploy by fkey
        :param fkey: The fkey of the deploy
        :return: The deploy object
        """
        try:
            async with httpx.AsyncClient(verify=VERIFY_HTTPS, timeout = httpx.Timeout(10800.0, connect=10.0, read=10800.0)) as client:
                response = await client.get(f"{APIGW}/Deploy/Fkey/{fkey}", headers=self.get_headers())
                logger.info(f"response {response.status_code} - {response.text}")
                if response.status_code == 200:
                    resp_json = response.json(cls=CustomJsonDecoder)
                    data = resp_json["data"]
                    deploy = Deploy(**data)
                    return deploy
                else:
                    logger.error(f"Error getting deploy by fkey: {fkey}")
                    return None
        except Exception as e:
            logger.error(f"Error getting deploy by fkey: {fkey}", exc_info=True)
            return None
            

    async def execute_run_ondemand(self, deploy_key: str, payload:dict|None = None) -> Run | None:
        """
        Execute a run on demand
        :param deploy_key: The deploy id to be executed
        :param payload: The payload to be used in the run
        :return: The run object
        """
        try:
            body = {
                "fKey": deploy_key,
                "payload": json.dumps(payload, cls=CustomJSONEncoder) if payload is not None else None
            }
            async with httpx.AsyncClient(verify=VERIFY_HTTPS, timeout = httpx.Timeout(10800.0, connect=10.0, read=10800.0)) as client:
                response = await client.post(f"{APIGW}/Run/ExecuteOnDemand", headers=self.get_headers(), json=body)
                if response.status_code == 201: 
                    json_res = response.json(cls=CustomJsonDecoder)
                    data = json_res["data"]
                    run = Run(**data)
                    return run
                else:
                    logger.error(f"Error executing run on demand, {response.status_code} - {response.text}")
                    return None
        except Exception as e:
            logger.error(f"Error executing run on demand - Deploy ID: {deploy_key}", exc_info=True)
            return None
    
    
    async def get_run(self, run_id: str) -> Run | None:
        """
        Get the result of a run
        :param run_id: The run id
        :return: The run object
        """
        try:
            async with httpx.AsyncClient(verify=VERIFY_HTTPS, timeout = httpx.Timeout(10800.0, connect=10.0, read=10800.0)) as client:
                response = await client.get(f"{APIGW}/Run/{run_id}", headers=self.get_headers())
                if response.status_code == 200: 
                    data = response.json()
                    run = Run(**data["data"])
                    return run
                else:
                    logger.error(f"Error getting run - Run ID: {run_id} - {response.text}")
                    return None
        except Exception as e:
            logger.error(f"Error getting run - Run ID: {run_id}", exc_info=True)
            return None
    
    
    async def get_deploy_runs(self, deploy_key:str, page = 1, page_size = 10) -> PaginationData[Run] | None:
        """
        Get runs
        :param deploy_key: The deploy key
        :return: The list of runs
        """
        try:
            async with httpx.AsyncClient(verify=VERIFY_HTTPS, timeout = httpx.Timeout(10800.0, connect=10.0, read=10800.0)) as client:
                response = await client.get(f"{APIGW}/Deploy/{deploy_key}/Run/History?Page={page}&PageSize={page_size}", headers=self.get_headers())
                if response.status_code == 200:
                    data = response.json()
                    runs = [Run(**run) for run in data["data"]["data"]]
                    total = data["data"]["total"]
                    return PaginationData(runs, page, page_size, total)
                elif response.status_code == 400:
                    logger.error(f"Error getting runs - Deploy Key: {deploy_key} - {response.text}")
                elif response.status_code == 404:
                    logger.error(f"Deploy {deploy_key} not found in your account, please check the deploy key or your credentials")
                return None
        
        except Exception as e:
            logger.error(f"Error getting runs - Deploy Key: {deploy_key}", exc_info=True)
            return None